@@ -34,7 +34,8 @@
 #include <fcntl.h>
 #include "fmm.h"
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtSetMemoryPolicy(HSAuint32 Node,
+HSAKMT_STATUS HSAKMTAPI hsaKmtSetMemoryPolicyCtx(HsaKFDContext *ctx,
+						  HSAuint32 Node,
 					      HSAuint32 DefaultPolicy,
 					      HSAuint32 AlternatePolicy,
 					      void *MemoryAddressAlternate,
@@ -86,7 +87,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtSetMemoryPolicy(HSAuint32 Node,
 	args.alternate_aperture_base = (uintptr_t) MemoryAddressAlternate;
 	args.alternate_aperture_size = MemorySizeInBytes;
 
-	int err = hsakmt_ioctl(hsakmt_kfd_fd, AMDKFD_IOC_SET_MEMORY_POLICY, &args);
+	int err = hsakmt_ioctl(ctx->fd, AMDKFD_IOC_SET_MEMORY_POLICY, &args);
 
 	return (err == -1) ? HSAKMT_STATUS_ERROR : HSAKMT_STATUS_SUCCESS;
 }
@@ -104,15 +105,17 @@ HSAuint32 hsakmt_PageSizeFromFlags(unsigned int pageSizeFlags)
 	}
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtAllocMemory(HSAuint32 PreferredNode,
+HSAKMT_STATUS HSAKMTAPI hsaKmtAllocMemoryCtx(HsaKFDContext *ctx,
+					  HSAuint32 PreferredNode,
 					  HSAuint64 SizeInBytes,
 					  HsaMemFlags MemFlags,
 					  void **MemoryAddress)
 {
-	return hsaKmtAllocMemoryAlign(PreferredNode, SizeInBytes, 0, MemFlags, MemoryAddress);
+	return hsaKmtAllocMemoryAlignCtx(ctx, PreferredNode, SizeInBytes, 0, MemFlags, MemoryAddress);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtAllocMemoryAlign(HSAuint32 PreferredNode,
+HSAKMT_STATUS HSAKMTAPI hsaKmtAllocMemoryAlignCtx(HsaKFDContext *ctx,
+					  HSAuint32 PreferredNode,
 					  HSAuint64 SizeInBytes,
 					  HSAuint64 Alignment,
 					  HsaMemFlags MemFlags,
@@ -160,7 +163,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtAllocMemoryAlign(HSAuint32 PreferredNode,
 			return HSAKMT_STATUS_NOT_IMPLEMENTED;
 		}
 
-		*MemoryAddress = hsakmt_fmm_allocate_scratch(gpu_id, *MemoryAddress, SizeInBytes);
+		*MemoryAddress = hsakmt_fmm_allocate_scratch(ctx, gpu_id, *MemoryAddress, SizeInBytes);
 
 		if (!(*MemoryAddress)) {
 			pr_err("[%s] failed to allocate %lu bytes from scratch\n",
@@ -183,7 +186,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtAllocMemoryAlign(HSAuint32 PreferredNode,
 		if (hsakmt_zfb_support && gpu_id && MemFlags.ui32.NonPaged == 1)
 			MemFlags.ui32.CoarseGrain = 1;
 
-		*MemoryAddress = hsakmt_fmm_allocate_host(gpu_id, MemFlags.ui32.GTTAccess ? 0 : PreferredNode,
+		*MemoryAddress = hsakmt_fmm_allocate_host(ctx, gpu_id, MemFlags.ui32.GTTAccess ? 0 : PreferredNode,
 						   *MemoryAddress, SizeInBytes, Alignment, MemFlags);
 
 		if (!(*MemoryAddress)) {
@@ -204,7 +207,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtAllocMemoryAlign(HSAuint32 PreferredNode,
 		return HSAKMT_STATUS_INVALID_PARAMETER;
 	}
 
-	*MemoryAddress = hsakmt_fmm_allocate_device(gpu_id, PreferredNode, *MemoryAddress,
+	*MemoryAddress = hsakmt_fmm_allocate_device(ctx, gpu_id, PreferredNode, *MemoryAddress,
 					     SizeInBytes, Alignment, MemFlags);
 
 	if (!(*MemoryAddress)) {
@@ -218,7 +221,8 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtAllocMemoryAlign(HSAuint32 PreferredNode,
 
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtFreeMemory(void *MemoryAddress,
+HSAKMT_STATUS HSAKMTAPI hsaKmtFreeMemoryCtx(HsaKFDContext *ctx,
+					 void *MemoryAddress,
 					 HSAuint64 SizeInBytes)
 {
 	CHECK_KFD_OPEN();
@@ -230,11 +234,12 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtFreeMemory(void *MemoryAddress,
 		return HSAKMT_STATUS_ERROR;
 	}
 
-	return hsakmt_fmm_release(MemoryAddress);
+	return hsakmt_fmm_release(ctx, MemoryAddress);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtAvailableMemory(HSAuint32 Node,
-					      HSAuint64 *AvailableBytes)
+HSAKMT_STATUS HSAKMTAPI hsaKmtAvailableMemoryCtx(HsaKFDContext *ctx,
+						  HSAuint32 Node,
+						  HSAuint64 *AvailableBytes)
 {
 	struct kfd_ioctl_get_available_memory_args args = {};
 	HSAKMT_STATUS result;
@@ -250,14 +255,15 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtAvailableMemory(HSAuint32 Node,
 		return result;
 	}
 
-	if (hsakmt_ioctl(hsakmt_kfd_fd, AMDKFD_IOC_AVAILABLE_MEMORY, &args))
+	if (hsakmt_ioctl(ctx->fd, AMDKFD_IOC_AVAILABLE_MEMORY, &args))
 		return HSAKMT_STATUS_ERROR;
 
 	*AvailableBytes = args.available;
 	return HSAKMT_STATUS_SUCCESS;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemory(void *MemoryAddress,
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemoryCtx(HsaKFDContext *ctx,
+					     void *MemoryAddress,
 					     HSAuint64 MemorySizeInBytes)
 {
 	CHECK_KFD_OPEN();
@@ -271,11 +277,13 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemory(void *MemoryAddress,
 	HsaMemFlags flags;
 	flags.ui32.CoarseGrain = 1;
 	flags.ui32.ExtendedCoherent = 0;
-	return hsakmt_fmm_register_memory(MemoryAddress, MemorySizeInBytes,
+	return hsakmt_fmm_register_memory(ctx,
+				   MemoryAddress, MemorySizeInBytes,
 				   NULL, 0, flags);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemoryToNodes(void *MemoryAddress,
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemoryToNodesCtx(HsaKFDContext *ctx,
+						    void *MemoryAddress,
 						    HSAuint64 MemorySizeInBytes,
 						    HSAuint64 NumberOfNodes,
 						    HSAuint32 *NodeArray)
@@ -299,7 +307,8 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemoryToNodes(void *MemoryAddress,
 		flags.ui32.CoarseGrain = 1;
 		flags.ui32.ExtendedCoherent = 0;
 
-		ret = hsakmt_fmm_register_memory(MemoryAddress, MemorySizeInBytes,
+		ret = hsakmt_fmm_register_memory(ctx,
+					  MemoryAddress, MemorySizeInBytes,
 					  gpu_id_array,
 					  NumberOfNodes*sizeof(uint32_t),
 					  flags);
@@ -310,7 +319,8 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemoryToNodes(void *MemoryAddress,
 	return ret;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemoryWithFlags(void *MemoryAddress,
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemoryWithFlagsCtx(HsaKFDContext *ctx,
+						    void *MemoryAddress,
 						    HSAuint64 MemorySizeInBytes,
 						    HsaMemFlags MemFlags)
 {
@@ -331,21 +341,24 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemoryWithFlags(void *MemoryAddress,
 		/* TODO: support mixed APU and dGPU configurations */
 		return HSAKMT_STATUS_NOT_SUPPORTED;
 
-	ret = hsakmt_fmm_register_memory(MemoryAddress, MemorySizeInBytes,
+	ret = hsakmt_fmm_register_memory(ctx,
+		MemoryAddress, MemorySizeInBytes,
 		NULL, 0, MemFlags);
 
 	return ret;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterGraphicsHandleToNodes(HSAuint64 GraphicsResourceHandle,
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterGraphicsHandleToNodesCtx(HsaKFDContext *ctx,
+							    HSAuint64 GraphicsResourceHandle,
 							    HsaGraphicsResourceInfo *GraphicsResourceInfo,
 							    HSAuint64 NumberOfNodes,
 							    HSAuint32 *NodeArray)
 {
        HSA_REGISTER_MEM_FLAGS regFlags;
        regFlags.Value = 0;
-        
-       return hsaKmtRegisterGraphicsHandleToNodesExt(GraphicsResourceHandle,
+
+       return hsaKmtRegisterGraphicsHandleToNodesExtCtx(ctx,
+						     GraphicsResourceHandle,
 						     GraphicsResourceInfo,
 						     NumberOfNodes,
 						     NodeArray,
@@ -353,7 +366,8 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterGraphicsHandleToNodes(HSAuint64 GraphicsRe
 
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterGraphicsHandleToNodesExt(HSAuint64 GraphicsResourceHandle,
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterGraphicsHandleToNodesExtCtx(HsaKFDContext *ctx,
+							       HSAuint64 GraphicsResourceHandle,
 							       HsaGraphicsResourceInfo *GraphicsResourceInfo,
 							       HSAuint64 NumberOfNodes,
 							       HSAuint32 *NodeArray,
@@ -371,7 +385,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterGraphicsHandleToNodesExt(HSAuint64 Graphic
 	}
 
 	if (ret == HSAKMT_STATUS_SUCCESS) {
-		ret = hsakmt_fmm_register_graphics_handle(
+		ret = hsakmt_fmm_register_graphics_handle(ctx,
 			GraphicsResourceHandle, GraphicsResourceInfo,
 			gpu_id_array, NumberOfNodes * sizeof(uint32_t), RegisterFlags);
 		if (ret != HSAKMT_STATUS_SUCCESS)
@@ -381,7 +395,8 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterGraphicsHandleToNodesExt(HSAuint64 Graphic
 	return ret;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtExportDMABufHandle(void *MemoryAddress,
+HSAKMT_STATUS HSAKMTAPI hsaKmtExportDMABufHandleCtx(HsaKFDContext *ctx,
+						 void *MemoryAddress,
 						 HSAuint64 MemorySizeInBytes,
 						 int *DMABufFd,
 						 HSAuint64 *Offset)
@@ -391,11 +406,13 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtExportDMABufHandle(void *MemoryAddress,
 
 	pr_debug("[%s] address %p\n", __func__, MemoryAddress);
 
-	return hsakmt_fmm_export_dma_buf_fd(MemoryAddress, MemorySizeInBytes,
+	return hsakmt_fmm_export_dma_buf_fd(ctx,
+				     MemoryAddress, MemorySizeInBytes,
 				     DMABufFd, Offset);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtShareMemory(void *MemoryAddress,
+HSAKMT_STATUS HSAKMTAPI hsaKmtShareMemoryCtx(HsaKFDContext *ctx,
+					  void *MemoryAddress,
 					  HSAuint64 SizeInBytes,
 					  HsaSharedMemoryHandle *SharedMemoryHandle)
 {
@@ -406,25 +423,28 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtShareMemory(void *MemoryAddress,
 	if (!SharedMemoryHandle)
 		return HSAKMT_STATUS_INVALID_PARAMETER;
 
-	return hsakmt_fmm_share_memory(MemoryAddress, SizeInBytes, SharedMemoryHandle);
+	return hsakmt_fmm_share_memory(ctx, MemoryAddress, SizeInBytes, SharedMemoryHandle);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterSharedHandle(const HsaSharedMemoryHandle *SharedMemoryHandle,
-						   void **MemoryAddress,
-						   HSAuint64 *SizeInBytes)
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterSharedHandleCtx(HsaKFDContext *ctx,
+					  const HsaSharedMemoryHandle *SharedMemoryHandle,
+					  void **MemoryAddress,
+					  HSAuint64 *SizeInBytes)
 {
 	CHECK_KFD_OPEN();
 
 	pr_debug("[%s] handle %p\n", __func__, SharedMemoryHandle);
 
-	return hsaKmtRegisterSharedHandleToNodes(SharedMemoryHandle,
+	return hsaKmtRegisterSharedHandleToNodesCtx(ctx,
+						 SharedMemoryHandle,
 						 MemoryAddress,
 						 SizeInBytes,
 						 0,
 						 NULL);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterSharedHandleToNodes(const HsaSharedMemoryHandle *SharedMemoryHandle,
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterSharedHandleToNodesCtx(HsaKFDContext *ctx,
+							  const HsaSharedMemoryHandle *SharedMemoryHandle,
 							  void **MemoryAddress,
 							  HSAuint64 *SizeInBytes,
 							  HSAuint64 NumberOfNodes,
@@ -447,7 +467,8 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterSharedHandleToNodes(const HsaSharedMemoryH
 			goto error;
 	}
 
-	ret = hsakmt_fmm_register_shared_memory(SharedMemoryHandle,
+	ret = hsakmt_fmm_register_shared_memory(ctx,
+					 SharedMemoryHandle,
 					 SizeInBytes,
 					 MemoryAddress,
 					 gpu_id_array,
@@ -487,17 +508,17 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtProcessVMWrite(HSAuint32 Pid,
 	return HSAKMT_STATUS_NOT_IMPLEMENTED;
 }
 
-
-HSAKMT_STATUS HSAKMTAPI hsaKmtDeregisterMemory(void *MemoryAddress)
+HSAKMT_STATUS HSAKMTAPI hsaKmtDeregisterMemoryCtx(HsaKFDContext *ctx, void *MemoryAddress)
 {
 	CHECK_KFD_OPEN();
 
 	pr_debug("[%s] address %p\n", __func__, MemoryAddress);
 
-	return hsakmt_fmm_deregister_memory(MemoryAddress);
+	return hsakmt_fmm_deregister_memory(ctx, MemoryAddress);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtMapMemoryToGPU(void *MemoryAddress,
+HSAKMT_STATUS HSAKMTAPI hsaKmtMapMemoryToGPUCtx(HsaKFDContext *ctx,
+					     void *MemoryAddress,
 					     HSAuint64 MemorySizeInBytes,
 					     HSAuint64 *AlternateVAGPU)
 {
@@ -513,10 +534,11 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtMapMemoryToGPU(void *MemoryAddress,
 	if (AlternateVAGPU)
 		*AlternateVAGPU = 0;
 
-	return hsakmt_fmm_map_to_gpu(MemoryAddress, MemorySizeInBytes, AlternateVAGPU);
+	return hsakmt_fmm_map_to_gpu(ctx, MemoryAddress, MemorySizeInBytes, AlternateVAGPU);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtMapMemoryToGPUNodes(void *MemoryAddress,
+HSAKMT_STATUS HSAKMTAPI hsaKmtMapMemoryToGPUNodesCtx(HsaKFDContext *ctx,
+						  void *MemoryAddress,
 						  HSAuint64 MemorySizeInBytes,
 						  HSAuint64 *AlternateVAGPU,
 						  HsaMemMapFlags MemMapFlags,
@@ -537,16 +559,15 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtMapMemoryToGPUNodes(void *MemoryAddress,
 	}
 
 	if (!hsakmt_is_dgpu && NumberOfNodes == 1)
-		return hsaKmtMapMemoryToGPU(MemoryAddress,
-				MemorySizeInBytes,
-				AlternateVAGPU);
+		return hsaKmtMapMemoryToGPUCtx(ctx, MemoryAddress,
+					MemorySizeInBytes, AlternateVAGPU);
 
 	ret = hsakmt_validate_nodeid_array(&gpu_id_array,
 				NumberOfNodes, NodeArray);
 	if (ret != HSAKMT_STATUS_SUCCESS)
 		return ret;
 
-	ret = hsakmt_fmm_map_to_gpu_nodes(MemoryAddress, MemorySizeInBytes,
+	ret = hsakmt_fmm_map_to_gpu_nodes(ctx, MemoryAddress, MemorySizeInBytes,
 		gpu_id_array, NumberOfNodes, AlternateVAGPU);
 
 	if (gpu_id_array)
@@ -555,7 +576,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtMapMemoryToGPUNodes(void *MemoryAddress,
 	return ret;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtUnmapMemoryToGPU(void *MemoryAddress)
+HSAKMT_STATUS HSAKMTAPI hsaKmtUnmapMemoryToGPUCtx(HsaKFDContext *ctx, void *MemoryAddress)
 {
 	CHECK_KFD_OPEN();
 
@@ -567,7 +588,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtUnmapMemoryToGPU(void *MemoryAddress)
 		return HSAKMT_STATUS_SUCCESS;
 	}
 
-	if (!hsakmt_fmm_unmap_from_gpu(MemoryAddress))
+	if (!hsakmt_fmm_unmap_from_gpu(ctx, MemoryAddress))
 		return HSAKMT_STATUS_SUCCESS;
 	else
 		return HSAKMT_STATUS_ERROR;
@@ -588,16 +609,16 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtMapGraphicHandle(HSAuint32 NodeId,
 	return HSAKMT_STATUS_NOT_IMPLEMENTED;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtUnmapGraphicHandle(HSAuint32 NodeId,
+HSAKMT_STATUS HSAKMTAPI hsaKmtUnmapGraphicHandleCtx(HsaKFDContext *ctx,
+						 HSAuint32 NodeId,
 						 HSAuint64 FlatMemoryAddress,
 						 HSAuint64 SizeInBytes)
 {
-	CHECK_KFD_OPEN();
-
-	return hsaKmtUnmapMemoryToGPU(PORT_UINT64_TO_VPTR(FlatMemoryAddress));
+	return hsaKmtUnmapMemoryToGPUCtx(ctx, PORT_UINT64_TO_VPTR(FlatMemoryAddress));
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtGetTileConfig(HSAuint32 NodeId, HsaGpuTileConfig *config)
+HSAKMT_STATUS HSAKMTAPI hsaKmtGetTileConfigCtx(HsaKFDContext *ctx,
+						 HSAuint32 NodeId, HsaGpuTileConfig *config)
 {
 	struct kfd_ioctl_get_tile_config_args args = {0};
 	uint32_t gpu_id;
@@ -623,7 +644,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtGetTileConfig(HSAuint32 NodeId, HsaGpuTileConfig *
 	args.num_tile_configs = config->NumTileConfigs;
 	args.num_macro_tile_configs = config->NumMacroTileConfigs;
 
-	if (hsakmt_ioctl(hsakmt_kfd_fd, AMDKFD_IOC_GET_TILE_CONFIG, &args) != 0)
+	if (hsakmt_ioctl(ctx->fd, AMDKFD_IOC_GET_TILE_CONFIG, &args) != 0)
 		return HSAKMT_STATUS_ERROR;
 
 	config->NumTileConfigs = args.num_tile_configs;
@@ -637,7 +658,8 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtGetTileConfig(HSAuint32 NodeId, HsaGpuTileConfig *
 	return HSAKMT_STATUS_SUCCESS;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtQueryPointerInfo(const void *Pointer,
+HSAKMT_STATUS HSAKMTAPI hsaKmtQueryPointerInfoCtx(HsaKFDContext *ctx,
+					       const void *Pointer,
 					       HsaPointerInfo *PointerInfo)
 {
 	CHECK_KFD_OPEN();
@@ -646,47 +668,264 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtQueryPointerInfo(const void *Pointer,
 
 	if (!PointerInfo)
 		return HSAKMT_STATUS_INVALID_PARAMETER;
-	return hsakmt_fmm_get_mem_info(Pointer, PointerInfo);
+	return hsakmt_fmm_get_mem_info(ctx, Pointer, PointerInfo);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtSetMemoryUserData(const void *Pointer,
+HSAKMT_STATUS HSAKMTAPI hsaKmtSetMemoryUserDataCtx(HsaKFDContext *ctx,
+						const void *Pointer,
 						void *UserData)
 {
 	CHECK_KFD_OPEN();
 
 	pr_debug("[%s] pointer %p\n", __func__, Pointer);
 
-	return hsakmt_fmm_set_mem_user_data(Pointer, UserData);
+	return hsakmt_fmm_set_mem_user_data(ctx, Pointer, UserData);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtReplaceAsanHeaderPage(void *addr)
+HSAKMT_STATUS HSAKMTAPI hsaKmtReplaceAsanHeaderPageCtx(HsaKFDContext *ctx, void *addr)
 {
 #ifdef SANITIZER_AMDGPU
 	pr_debug("[%s] address %p\n", __func__, addr);
 	CHECK_KFD_OPEN();
 
-	return hsakmt_fmm_replace_asan_header_page(addr);
+	return hsakmt_fmm_replace_asan_header_page(ctx, addr);
 #else
 	return HSAKMT_STATUS_NOT_SUPPORTED;
 #endif
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtReturnAsanHeaderPage(void *addr)
+HSAKMT_STATUS HSAKMTAPI hsaKmtReturnAsanHeaderPageCtx(HsaKFDContext *ctx, void *addr)
 {
 #ifdef SANITIZER_AMDGPU
 	pr_debug("[%s] address %p\n", __func__, addr);
 	CHECK_KFD_OPEN();
 
-	return hsakmt_fmm_return_asan_header_page(addr);
+	return hsakmt_fmm_return_asan_header_page(ctx, addr);
 #else
 	return HSAKMT_STATUS_NOT_SUPPORTED;
 #endif
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtGetAMDGPUDeviceHandle( HSAuint32 NodeId,
+HSAKMT_STATUS HSAKMTAPI hsaKmtGetAMDGPUDeviceHandleCtx(HsaKFDContext *ctx,
+						HSAuint32 NodeId,
 						HsaAMDGPUDeviceHandle   *DeviceHandle)
 {
 	CHECK_KFD_OPEN();
 
-	return hsakmt_fmm_get_amdgpu_device_handle(NodeId, DeviceHandle);
+	return hsakmt_fmm_get_amdgpu_device_handle(ctx, NodeId, DeviceHandle);
+}
+
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtSetMemoryPolicy(HSAuint32 Node,
+					      HSAuint32 DefaultPolicy,
+					      HSAuint32 AlternatePolicy,
+					      void *MemoryAddressAlternate,
+					      HSAuint64 MemorySizeInBytes)
+{
+	return hsaKmtSetMemoryPolicyCtx(&hsakmt_primary_kfd_ctx, Node,
+					  DefaultPolicy, AlternatePolicy,
+					  MemoryAddressAlternate, MemorySizeInBytes);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtAllocMemory(HSAuint32 PreferredNode,
+					  HSAuint64 SizeInBytes,
+					  HsaMemFlags MemFlags,
+					  void **MemoryAddress)
+{
+	return hsaKmtAllocMemoryAlign(PreferredNode, SizeInBytes, 0, MemFlags, MemoryAddress);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtAllocMemoryAlign(HSAuint32 PreferredNode,
+					  HSAuint64 SizeInBytes,
+					  HSAuint64 Alignment,
+					  HsaMemFlags MemFlags,
+					  void **MemoryAddress)
+{
+	return hsaKmtAllocMemoryAlignCtx(&hsakmt_primary_kfd_ctx, PreferredNode,
+					  SizeInBytes, Alignment, MemFlags, MemoryAddress);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtFreeMemory(void *MemoryAddress,
+					  HSAuint64 SizeInBytes)
+{
+	return hsaKmtFreeMemoryCtx(&hsakmt_primary_kfd_ctx, MemoryAddress, SizeInBytes);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtAvailableMemory(HSAuint32 Node,
+					  HSAuint64 *AvailableBytes)
+{
+	return hsaKmtAvailableMemoryCtx(&hsakmt_primary_kfd_ctx, Node, AvailableBytes);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemory(void *MemoryAddress,
+					      HSAuint64 MemorySizeInBytes)
+{
+	return hsaKmtRegisterMemoryCtx(&hsakmt_primary_kfd_ctx, MemoryAddress, MemorySizeInBytes);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemoryToNodes(void *MemoryAddress,
+					      HSAuint64 MemorySizeInBytes,
+					      HSAuint64 NumberOfNodes,
+					      HSAuint32 *NodeArray)
+{
+	return hsaKmtRegisterMemoryToNodesCtx(&hsakmt_primary_kfd_ctx,
+					      MemoryAddress, MemorySizeInBytes,
+					      NumberOfNodes, NodeArray);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterMemoryWithFlags(void *MemoryAddress,
+					      HSAuint64 MemorySizeInBytes,
+					      HsaMemFlags MemFlags)
+{
+	return hsaKmtRegisterMemoryWithFlagsCtx(&hsakmt_primary_kfd_ctx,
+					      MemoryAddress, MemorySizeInBytes, MemFlags);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterGraphicsHandleToNodes(HSAuint64 GraphicsResourceHandle,
+							    HsaGraphicsResourceInfo *GraphicsResourceInfo,
+							    HSAuint64 NumberOfNodes,
+							    HSAuint32 *NodeArray)
+{
+       HSA_REGISTER_MEM_FLAGS regFlags;
+       regFlags.Value = 0;
+
+       return hsaKmtRegisterGraphicsHandleToNodesExt(GraphicsResourceHandle,
+						     GraphicsResourceInfo,
+						     NumberOfNodes,
+						     NodeArray,
+						     regFlags);
+
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterGraphicsHandleToNodesExt(HSAuint64 GraphicsResourceHandle,
+							       HsaGraphicsResourceInfo *GraphicsResourceInfo,
+							       HSAuint64 NumberOfNodes,
+							       HSAuint32 *NodeArray,
+							       HSA_REGISTER_MEM_FLAGS RegisterFlags)
+{
+	return hsaKmtRegisterGraphicsHandleToNodesExtCtx(&hsakmt_primary_kfd_ctx,
+							       GraphicsResourceHandle,
+							       GraphicsResourceInfo,
+							       NumberOfNodes,
+							       NodeArray,
+							       RegisterFlags);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtExportDMABufHandle(void *MemoryAddress,
+						 HSAuint64 MemorySizeInBytes,
+						 int *DMABufFd,
+						 HSAuint64 *Offset)
+{
+	return hsaKmtExportDMABufHandleCtx(&hsakmt_primary_kfd_ctx,
+					     MemoryAddress, MemorySizeInBytes,
+					     DMABufFd, Offset);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtShareMemory(void *MemoryAddress,
+					  HSAuint64 SizeInBytes,
+					  HsaSharedMemoryHandle *SharedMemoryHandle)
+{
+	return hsaKmtShareMemoryCtx(&hsakmt_primary_kfd_ctx,
+				     MemoryAddress, SizeInBytes, SharedMemoryHandle);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterSharedHandle(
+					  const HsaSharedMemoryHandle *SharedMemoryHandle,
+					  void **MemoryAddress,
+					  HSAuint64 *SizeInBytes)
+{
+	CHECK_KFD_OPEN();
+
+	pr_debug("[%s] handle %p\n", __func__, SharedMemoryHandle);
+
+	return hsaKmtRegisterSharedHandleToNodes(SharedMemoryHandle,
+						     MemoryAddress,
+						     SizeInBytes,
+						     0,
+						     NULL);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtRegisterSharedHandleToNodes(const HsaSharedMemoryHandle *SharedMemoryHandle,
+							  void **MemoryAddress,
+							  HSAuint64 *SizeInBytes,
+							  HSAuint64 NumberOfNodes,
+							  HSAuint32 *NodeArray)
+{
+	return hsaKmtRegisterSharedHandleToNodesCtx(&hsakmt_primary_kfd_ctx,
+						     SharedMemoryHandle,
+						     MemoryAddress,
+						     SizeInBytes,
+						     NumberOfNodes,
+						     NodeArray);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtDeregisterMemory(void *MemoryAddress)
+{
+	return hsaKmtDeregisterMemoryCtx(&hsakmt_primary_kfd_ctx, MemoryAddress);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtMapMemoryToGPU(void *MemoryAddress,
+					  HSAuint64 MemorySizeInBytes,
+					  HSAuint64 *AlternateVAGPU)
+{
+	return hsaKmtMapMemoryToGPUCtx(&hsakmt_primary_kfd_ctx, MemoryAddress, MemorySizeInBytes, AlternateVAGPU);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtMapMemoryToGPUNodes(
+					  void *MemoryAddress,
+					  HSAuint64 MemorySizeInBytes,
+					  HSAuint64 *AlternateVAGPU,
+					  HsaMemMapFlags MemMapFlags,
+					  HSAuint64 NumberOfNodes,
+					  HSAuint32 *NodeArray)
+{
+	return hsaKmtMapMemoryToGPUNodesCtx(&hsakmt_primary_kfd_ctx, MemoryAddress,
+				MemorySizeInBytes, AlternateVAGPU, MemMapFlags, NumberOfNodes, NodeArray);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtUnmapMemoryToGPU(void *MemoryAddress)
+{
+	return hsaKmtUnmapMemoryToGPUCtx(&hsakmt_primary_kfd_ctx, MemoryAddress);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtUnmapGraphicHandle(HSAuint32 NodeId,
+					  HSAuint64 FlatMemoryAddress,
+					  HSAuint64 SizeInBytes)
+{
+	return hsaKmtUnmapMemoryToGPU(PORT_UINT64_TO_VPTR(FlatMemoryAddress));
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtGetTileConfig(HSAuint32 NodeId, HsaGpuTileConfig *config)
+{
+	return hsaKmtGetTileConfigCtx(&hsakmt_primary_kfd_ctx, NodeId, config);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtQueryPointerInfo(const void *Pointer,
+					       HsaPointerInfo *PointerInfo)
+{
+	return hsaKmtQueryPointerInfoCtx(&hsakmt_primary_kfd_ctx, Pointer, PointerInfo);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtSetMemoryUserData(const void *Pointer,
+						void *UserData)
+{
+	return hsaKmtSetMemoryUserDataCtx(&hsakmt_primary_kfd_ctx, Pointer, UserData);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtReplaceAsanHeaderPage(void *addr)
+{
+	return hsaKmtReplaceAsanHeaderPageCtx(&hsakmt_primary_kfd_ctx, addr);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtReturnAsanHeaderPage(void *addr)
+{
+	return hsaKmtReturnAsanHeaderPageCtx(&hsakmt_primary_kfd_ctx, addr);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtGetAMDGPUDeviceHandle(HSAuint32 NodeId,
+						HsaAMDGPUDeviceHandle   *DeviceHandle)
+{
+	CHECK_KFD_OPEN();
+
+	return hsaKmtGetAMDGPUDeviceHandleCtx(&hsakmt_primary_kfd_ctx, NodeId, DeviceHandle);
 }
