@@ -35,6 +35,7 @@
 #include <sys/mman.h>
 #include <fcntl.h>
 #include <errno.h>
+#include <assert.h>
 
 /* 1024 doorbells, 4 or 8 bytes each doorbell depending on ASIC generation */
 #define DOORBELL_SIZE(gfxv)	(((gfxv) >= 0x90000) ? 8 : 4)
@@ -80,8 +81,28 @@ struct process_doorbells {
 	pthread_mutex_t mutex;
 };
 
-static unsigned int num_doorbells;
-static struct process_doorbells *doorbells;
+struct hsa_kfd_queue_context
+{
+	unsigned int num_doorbells;
+	struct process_doorbells *doorbells;
+};
+
+struct hsa_kfd_queue_context *hsakmt_kfdcontext_get_queue_context(HsaKFDContext *ctx)
+{
+	assert(ctx);
+
+	if (ctx->queue_context)
+		return ctx->queue_context;
+
+	ctx->queue_context = calloc(1, sizeof(struct hsa_kfd_queue_context));
+	if (!ctx->queue_context) {
+		pr_err("Alloc memory failed for struct hsa_kfd_queue_context size %zu\n",
+				 sizeof(struct hsa_kfd_queue_context));
+		return NULL;
+	}
+
+	return ctx->queue_context;
+}
 
 uint32_t hsakmt_get_vgpr_size_per_cu(uint32_t gfxv)
 {
@@ -103,26 +124,27 @@ uint32_t hsakmt_get_vgpr_size_per_cu(uint32_t gfxv)
 	return vgpr_size;
 }
 
-HSAKMT_STATUS hsakmt_init_process_doorbells(unsigned int NumNodes)
+HSAKMT_STATUS hsakmt_init_process_doorbells(HsaKFDContext *ctx, unsigned int NumNodes)
 {
 	unsigned int i;
 	HSAKMT_STATUS ret = HSAKMT_STATUS_SUCCESS;
+	struct hsa_kfd_queue_context *queue_ctx = hsakmt_kfdcontext_get_queue_context(ctx);
 
-	/* doorbells[] is accessed using Topology NodeId. This means doorbells[0],
+	/* queue_ctx->doorbells[] is accessed using Topology NodeId. This means doorbells[0],
 	 * which corresponds to CPU only Node, might not be used
 	 */
-	doorbells = malloc(NumNodes * sizeof(struct process_doorbells));
-	if (!doorbells)
+	queue_ctx->doorbells = malloc(NumNodes * sizeof(struct process_doorbells));
+	if (!queue_ctx->doorbells)
 		return HSAKMT_STATUS_NO_MEMORY;
 
 	for (i = 0; i < NumNodes; i++) {
-		doorbells[i].use_gpuvm = false;
-		doorbells[i].size = 0;
-		doorbells[i].mapping = NULL;
-		pthread_mutex_init(&doorbells[i].mutex, NULL);
+		queue_ctx->doorbells[i].use_gpuvm = false;
+		queue_ctx->doorbells[i].size = 0;
+		queue_ctx->doorbells[i].mapping = NULL;
+		pthread_mutex_init(&queue_ctx->doorbells[i].mutex, NULL);
 	}
 
-	num_doorbells = NumNodes;
+	queue_ctx->num_doorbells = NumNodes;
 
 	return ret;
 }
@@ -145,94 +167,105 @@ static void get_doorbell_map_info(uint32_t node_id,
 	return;
 }
 
-void hsakmt_destroy_process_doorbells(void)
+void hsakmt_destroy_process_doorbells(HsaKFDContext *ctx)
 {
 	unsigned int i;
+	struct hsa_kfd_queue_context *queue_ctx = hsakmt_kfdcontext_get_queue_context(ctx);
+	struct process_doorbells *doorbells = queue_ctx->doorbells;
 
 	if (!doorbells)
 		return;
 
-	for (i = 0; i < num_doorbells; i++) {
+	for (i = 0; i < queue_ctx->num_doorbells; i++) {
 		if (!doorbells[i].size)
 			continue;
 
 		if (doorbells[i].use_gpuvm) {
-			hsakmt_fmm_unmap_from_gpu(doorbells[i].mapping);
-			hsakmt_fmm_release(doorbells[i].mapping);
+			hsakmt_fmm_unmap_from_gpu(ctx, doorbells[i].mapping);
+			hsakmt_fmm_release(ctx, doorbells[i].mapping);
 		} else
 			munmap(doorbells[i].mapping, doorbells[i].size);
 	}
 
 	free(doorbells);
-	doorbells = NULL;
-	num_doorbells = 0;
+	queue_ctx->doorbells = NULL;
+	queue_ctx->num_doorbells = 0;
 }
 
 /* This is a special funcion that should be called only from the child process
  * after a fork(). This will clear doorbells duplicated from the parent.
  */
-void hsakmt_clear_process_doorbells(void)
+void hsakmt_clear_process_doorbells(HsaKFDContext *ctx)
 {
 	unsigned int i;
+	struct hsa_kfd_queue_context *queue_ctx = hsakmt_kfdcontext_get_queue_context(ctx);
 
-	if (!doorbells)
+	if (!queue_ctx->doorbells)
 		return;
 
-	for (i = 0; i < num_doorbells; i++) {
-		if (!doorbells[i].size)
+	for (i = 0; i < queue_ctx->num_doorbells; i++) {
+		if (!queue_ctx->doorbells[i].size)
 			continue;
 
-		if (!doorbells[i].use_gpuvm)
-			munmap(doorbells[i].mapping, doorbells[i].size);
+		if (!queue_ctx->doorbells[i].use_gpuvm)
+			munmap(queue_ctx->doorbells[i].mapping, queue_ctx->doorbells[i].size);
 	}
 
-	free(doorbells);
-	doorbells = NULL;
-	num_doorbells = 0;
+	free(queue_ctx->doorbells);
+	queue_ctx->doorbells = NULL;
+	queue_ctx->num_doorbells = 0;
 }
 
-static HSAKMT_STATUS map_doorbell_apu(HSAuint32 NodeId, HSAuint32 gpu_id,
+static HSAKMT_STATUS map_doorbell_apu(HsaKFDContext *ctx,
+					  HSAuint32 NodeId, HSAuint32 gpu_id,
 				      HSAuint64 doorbell_mmap_offset)
 {
 	void *ptr;
+	struct hsa_kfd_queue_context *queue_ctx = hsakmt_kfdcontext_get_queue_context(ctx);
 
-	ptr = mmap(0, doorbells[NodeId].size, PROT_READ|PROT_WRITE,
-		   MAP_SHARED, hsakmt_kfd_fd, doorbell_mmap_offset);
+	ptr = mmap(0, queue_ctx->doorbells[NodeId].size, PROT_READ|PROT_WRITE,
+		   MAP_SHARED, ctx->fd, doorbell_mmap_offset);
 
 	if (ptr == MAP_FAILED)
 		return HSAKMT_STATUS_ERROR;
 
-	doorbells[NodeId].mapping = ptr;
+	queue_ctx->doorbells[NodeId].mapping = ptr;
 
 	return HSAKMT_STATUS_SUCCESS;
 }
 
-static HSAKMT_STATUS map_doorbell_dgpu(HSAuint32 NodeId, HSAuint32 gpu_id,
+static HSAKMT_STATUS map_doorbell_dgpu(HsaKFDContext *ctx,
+					   HSAuint32 NodeId, HSAuint32 gpu_id,
 				       HSAuint64 doorbell_mmap_offset)
 {
 	void *ptr;
+	struct hsa_kfd_queue_context *queue_ctx = hsakmt_kfdcontext_get_queue_context(ctx);
 
-	ptr = hsakmt_fmm_allocate_doorbell(gpu_id, doorbells[NodeId].size,
+	ptr = hsakmt_fmm_allocate_doorbell(ctx,
+				gpu_id, queue_ctx->doorbells[NodeId].size,
 				doorbell_mmap_offset);
 
 	if (!ptr)
 		return HSAKMT_STATUS_ERROR;
 
 	/* map for GPU access */
-	if (hsakmt_fmm_map_to_gpu(ptr, doorbells[NodeId].size, NULL)) {
-		hsakmt_fmm_release(ptr);
+	if (hsakmt_fmm_map_to_gpu(ctx, ptr, queue_ctx->doorbells[NodeId].size, NULL)) {
+		hsakmt_fmm_release(ctx, ptr);
 		return HSAKMT_STATUS_ERROR;
 	}
 
-	doorbells[NodeId].mapping = ptr;
+	queue_ctx->doorbells[NodeId].mapping = ptr;
 
 	return HSAKMT_STATUS_SUCCESS;
 }
 
-static HSAKMT_STATUS map_doorbell(HSAuint32 NodeId, HSAuint32 gpu_id,
+static HSAKMT_STATUS map_doorbell(HsaKFDContext *ctx,
+				  HSAuint32 NodeId, HSAuint32 gpu_id,
 				  HSAuint64 doorbell_mmap_offset)
 {
 	HSAKMT_STATUS status = HSAKMT_STATUS_SUCCESS;
+	struct hsa_kfd_queue_context *queue_ctx = hsakmt_kfdcontext_get_queue_context(ctx);
+	struct process_doorbells *doorbells = queue_ctx->doorbells;
 
 	pthread_mutex_lock(&doorbells[NodeId].mutex);
 	if (doorbells[NodeId].size) {
@@ -243,16 +276,16 @@ static HSAKMT_STATUS map_doorbell(HSAuint32 NodeId, HSAuint32 gpu_id,
 	get_doorbell_map_info(NodeId, &doorbells[NodeId]);
 
 	if (doorbells[NodeId].use_gpuvm) {
-		status = map_doorbell_dgpu(NodeId, gpu_id, doorbell_mmap_offset);
+		status = map_doorbell_dgpu(ctx, NodeId, gpu_id, doorbell_mmap_offset);
 		if (status != HSAKMT_STATUS_SUCCESS) {
 			/* Fall back to the old method if KFD doesn't
 			 * support doorbells in GPUVM
 			 */
 			doorbells[NodeId].use_gpuvm = false;
-			status = map_doorbell_apu(NodeId, gpu_id, doorbell_mmap_offset);
+			status = map_doorbell_apu(ctx, NodeId, gpu_id, doorbell_mmap_offset);
 		}
 	} else
-		status = map_doorbell_apu(NodeId, gpu_id, doorbell_mmap_offset);
+		status = map_doorbell_apu(ctx, NodeId, gpu_id, doorbell_mmap_offset);
 
 	if (status != HSAKMT_STATUS_SUCCESS)
 		doorbells[NodeId].size = 0;
@@ -280,13 +313,13 @@ static void *allocate_exec_aligned_memory_cpu(uint32_t size)
 }
 
 /* The bool return indicate whether the queue needs a context-save-restore area*/
-static bool update_ctx_save_restore_size(uint32_t nodeid, struct queue *q)
+static bool update_ctx_save_restore_size(HsaKFDContext *ctx, uint32_t nodeid, struct queue *q)
 {
 	HsaNodeProperties node;
 
 	if (q->gfxv < GFX_VERSION_CARRIZO)
 		return false;
-	if (hsaKmtGetNodeProperties(nodeid, &node))
+	if (hsaKmtGetNodePropertiesCtx(ctx, nodeid, &node))
 		return false;
 	if (node.NumFComputeCores && node.NumSIMDPerCU) {
 		uint32_t ctl_stack_size, wg_data_size;
@@ -317,7 +350,8 @@ static bool update_ctx_save_restore_size(uint32_t nodeid, struct queue *q)
 	return false;
 }
 
-void *hsakmt_allocate_exec_aligned_memory_gpu(uint32_t size, uint32_t align, uint32_t gpu_id,
+void *hsakmt_allocate_exec_aligned_memory_gpu(HsaKFDContext *ctx,
+					   uint32_t size, uint32_t align, uint32_t gpu_id,
 				       uint32_t NodeId, bool nonPaged,
 				       bool DeviceLocal,
 				       bool Uncached)
@@ -338,7 +372,7 @@ void *hsakmt_allocate_exec_aligned_memory_gpu(uint32_t size, uint32_t align, uin
 	size = ALIGN_UP(size, align);
 
 	if (DeviceLocal && !hsakmt_zfb_support)
-		mem = hsakmt_fmm_allocate_device(gpu_id, NodeId, mem, size, 0, flags);
+		mem = hsakmt_fmm_allocate_device(ctx, gpu_id, NodeId, mem, size, 0, flags);
 	else {
 		/* VRAM under ZFB mode should be supported here without any
 		 * additional code
@@ -353,7 +387,7 @@ void *hsakmt_allocate_exec_aligned_memory_gpu(uint32_t size, uint32_t align, uin
 				cpu_id = 0;
 			}
 		}
-		mem = hsakmt_fmm_allocate_host(gpu_id, cpu_id, mem, size, 0, flags);
+		mem = hsakmt_fmm_allocate_host(ctx, gpu_id, cpu_id, mem, size, 0, flags);
 	}
 
 	if (!mem) {
@@ -367,35 +401,36 @@ void *hsakmt_allocate_exec_aligned_memory_gpu(uint32_t size, uint32_t align, uin
 		HsaMemMapFlags map_flags = {0};
 		HSAKMT_STATUS result;
 
-		result = hsaKmtMapMemoryToGPUNodes(mem, size, &gpu_va, map_flags, 1, nodes_array);
+		result = hsaKmtMapMemoryToGPUNodesCtx(ctx, mem, size, &gpu_va, map_flags, 1, nodes_array);
 		if (result != HSAKMT_STATUS_SUCCESS) {
-			hsaKmtFreeMemory(mem, size);
+			hsaKmtFreeMemoryCtx(ctx, mem, size);
 			return NULL;
 		}
 
 		return mem;
 	}
 
-	if (hsaKmtMapMemoryToGPU(mem, size, &gpu_va) != HSAKMT_STATUS_SUCCESS) {
-		hsaKmtFreeMemory(mem, size);
+	if (hsaKmtMapMemoryToGPUCtx(ctx, mem, size, &gpu_va) != HSAKMT_STATUS_SUCCESS) {
+		hsaKmtFreeMemoryCtx(ctx, mem, size);
 		return NULL;
 	}
 
 	return mem;
 }
 
-void hsakmt_free_exec_aligned_memory_gpu(void *addr, uint32_t size, uint32_t align)
+void hsakmt_free_exec_aligned_memory_gpu(HsaKFDContext *ctx, void *addr, uint32_t size, uint32_t align)
 {
 	size = ALIGN_UP(size, align);
 
-	if (hsaKmtUnmapMemoryToGPU(addr) == HSAKMT_STATUS_SUCCESS)
-		hsaKmtFreeMemory(addr, size);
+	if (hsaKmtUnmapMemoryToGPUCtx(ctx, addr) == HSAKMT_STATUS_SUCCESS)
+		hsaKmtFreeMemoryCtx(ctx, addr, size);
 }
 
 /*
  * Allocates memory aligned to sysconf(_SC_PAGESIZE)
  */
-static void *allocate_exec_aligned_memory(uint32_t size,
+static void *allocate_exec_aligned_memory(HsaKFDContext *ctx,
+					  uint32_t size,
 					  bool use_ats,
 					  uint32_t gpu_id,
 					  uint32_t NodeId,
@@ -404,17 +439,19 @@ static void *allocate_exec_aligned_memory(uint32_t size,
 					  bool Uncached)
 {
 	if (!use_ats)
-		return hsakmt_allocate_exec_aligned_memory_gpu(size, PAGE_SIZE, gpu_id, NodeId,
+		return hsakmt_allocate_exec_aligned_memory_gpu(ctx,
+							size, PAGE_SIZE, gpu_id, NodeId,
 							nonPaged, DeviceLocal,
 							Uncached);
 	return allocate_exec_aligned_memory_cpu(size);
 }
 
-static void free_exec_aligned_memory(void *addr, uint32_t size, uint32_t align,
+static void free_exec_aligned_memory(HsaKFDContext *ctx,
+				     void *addr, uint32_t size, uint32_t align,
 				     bool use_ats)
 {
 	if (!use_ats)
-		hsakmt_free_exec_aligned_memory_gpu(addr, size, align);
+		hsakmt_free_exec_aligned_memory_gpu(ctx, addr, size, align);
 	else
 		munmap(addr, size);
 }
@@ -455,20 +492,20 @@ static HSAKMT_STATUS register_svm_range(void *mem, uint32_t size,
 	return hsaKmtSVMSetAttr(mem, size, nattr, attrs);
 }
 
-static void free_queue(struct queue *q)
+static void free_queue(HsaKFDContext *ctx, struct queue *q)
 {
 	if (q->eop_buffer)
-		free_exec_aligned_memory(q->eop_buffer,
+		free_exec_aligned_memory(ctx, q->eop_buffer,
 					 q->eop_buffer_size,
 					 PAGE_SIZE, q->use_ats);
 	if (q->unified_ctx_save_restore)
 		munmap(q->ctx_save_restore, q->total_mem_alloc_size);
 	else if (q->ctx_save_restore)
-		free_exec_aligned_memory(q->ctx_save_restore,
+		free_exec_aligned_memory(ctx, q->ctx_save_restore,
 					 q->total_mem_alloc_size,
 					 PAGE_SIZE, q->use_ats);
 
-	free_exec_aligned_memory((void *)q, sizeof(*q), PAGE_SIZE, q->use_ats);
+	free_exec_aligned_memory(ctx, (void *)q, sizeof(*q), PAGE_SIZE, q->use_ats);
 }
 
 static inline void fill_cwsr_header(struct queue *q, void *addr,
@@ -489,7 +526,8 @@ static inline void fill_cwsr_header(struct queue *q, void *addr,
 	}
 }
 
-static int handle_concrete_asic(struct queue *q,
+static int handle_concrete_asic(HsaKFDContext *ctx,
+				struct queue *q,
 				struct kfd_ioctl_create_queue_args *args,
 				uint32_t gpu_id,
 				uint32_t NodeId,
@@ -504,7 +542,8 @@ static int handle_concrete_asic(struct queue *q,
 
 	if (q->eop_buffer_size > 0) {
 		pr_info("Allocating VRAM for EOP\n");
-		q->eop_buffer = allocate_exec_aligned_memory(q->eop_buffer_size,
+		q->eop_buffer = allocate_exec_aligned_memory(ctx,
+				q->eop_buffer_size,
 				q->use_ats, gpu_id,
 				NodeId, true, true, /* Unused for VRAM */false);
 		if (!q->eop_buffer)
@@ -514,12 +553,12 @@ static int handle_concrete_asic(struct queue *q,
 		args->eop_buffer_size = q->eop_buffer_size;
 	}
 
-	ret = update_ctx_save_restore_size(NodeId, q);
+	ret = update_ctx_save_restore_size(ctx, NodeId, q);
 
 	if (ret) {
 		HsaNodeProperties node;
 
-		if (hsaKmtGetNodeProperties(NodeId, &node))
+		if (hsaKmtGetNodePropertiesCtx(ctx, NodeId, &node))
 			return HSAKMT_STATUS_ERROR;
 
 		args->ctx_save_restore_size = q->ctx_save_restore_size;
@@ -569,7 +608,7 @@ static int handle_concrete_asic(struct queue *q,
 		}
 
 		if (!q->unified_ctx_save_restore) {
-			q->ctx_save_restore = allocate_exec_aligned_memory(
+			q->ctx_save_restore = allocate_exec_aligned_memory(ctx,
 							q->total_mem_alloc_size,
 							q->use_ats, gpu_id, NodeId,
 							false, false, false);
@@ -592,24 +631,26 @@ static int handle_concrete_asic(struct queue *q,
  */
 static uint32_t priority_map[] = {0, 3, 5, 7, 9, 11, 15};
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueue(HSAuint32 NodeId,
-					  HSA_QUEUE_TYPE Type,
-					  HSAuint32 QueuePercentage,
-					  HSA_QUEUE_PRIORITY Priority,
-					  void *QueueAddress,
-					  HSAuint64 QueueSizeInBytes,
-					  HsaEvent *Event,
-					  HsaQueueResource *QueueResource)
+HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueueCtx(HsaKFDContext *ctx,
+						 HSAuint32 NodeId,
+						 HSA_QUEUE_TYPE Type,
+						 HSAuint32 QueuePercentage,
+						 HSA_QUEUE_PRIORITY Priority,
+						 void *QueueAddress,
+						 HSAuint64 QueueSizeInBytes,
+						 HsaEvent *Event,
+						 HsaQueueResource *QueueResource)
 {
 	if (Type == HSA_QUEUE_SDMA_BY_ENG_ID)
 		return HSAKMT_STATUS_ERROR;
 
-	return hsaKmtCreateQueueExt(NodeId, Type, QueuePercentage, Priority, 0,
+	return hsaKmtCreateQueueExtCtx(ctx, NodeId, Type, QueuePercentage, Priority, 0,
 				    QueueAddress, QueueSizeInBytes, Event,
 				    QueueResource);
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueueExt(HSAuint32 NodeId,
+HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueueExtCtx(HsaKFDContext *ctx,
+						 HSAuint32 NodeId,
 					     HSA_QUEUE_TYPE Type,
 					     HSAuint32 QueuePercentage,
 					     HSA_QUEUE_PRIORITY Priority,
@@ -629,6 +670,8 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueueExt(HSAuint32 NodeId,
 
 	CHECK_KFD_OPEN();
 
+	struct hsa_kfd_queue_context *queue_ctx = hsakmt_kfdcontext_get_queue_context(ctx);
+
 	if (Priority < HSA_QUEUE_PRIORITY_MINIMUM ||
 		Priority > HSA_QUEUE_PRIORITY_MAXIMUM)
 		return HSAKMT_STATUS_INVALID_PARAMETER;
@@ -637,7 +680,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueueExt(HSAuint32 NodeId,
 	if (result != HSAKMT_STATUS_SUCCESS)
 		return result;
 
-	struct queue *q = allocate_exec_aligned_memory(sizeof(*q),
+	struct queue *q = allocate_exec_aligned_memory(ctx, sizeof(*q),
 			false, gpu_id, NodeId, true, false, true);
 	if (!q)
 		return HSAKMT_STATUS_NO_MEMORY;
@@ -657,7 +700,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueueExt(HSAuint32 NodeId,
 	/* By default, CUs are all turned on. Initialize cu_mask to '1
 	 * for all CU bits.
 	 */
-	if (hsaKmtGetNodeProperties(NodeId, &props))
+	if (hsaKmtGetNodePropertiesCtx(ctx, NodeId, &props))
 		q->cu_mask_count = 0;
 	else {
 		cu_num = props.NumFComputeCores / props.NumSIMDPerCU;
@@ -696,9 +739,9 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueueExt(HSAuint32 NodeId,
 		QueueResource->QueueWptrValue = (uintptr_t)&q->wptr;
 	}
 
-	err = handle_concrete_asic(q, &args, gpu_id, NodeId, Event, QueueResource->ErrorReason);
+	err = handle_concrete_asic(ctx, q, &args, gpu_id, NodeId, Event, QueueResource->ErrorReason);
 	if (err != HSAKMT_STATUS_SUCCESS) {
-		free_queue(q);
+		free_queue(ctx, q);
 		return err;
 	}
 
@@ -710,10 +753,10 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueueExt(HSAuint32 NodeId,
 	args.queue_priority = priority_map[Priority+3];
 	args.sdma_engine_id = SdmaEngineId;
 
-	err = hsakmt_ioctl(hsakmt_kfd_fd, AMDKFD_IOC_CREATE_QUEUE, &args);
+	err = hsakmt_ioctl(ctx->fd, AMDKFD_IOC_CREATE_QUEUE, &args);
 
 	if (err == -1) {
-		free_queue(q);
+		free_queue(ctx, q);
 		return HSAKMT_STATUS_ERROR;
 	}
 
@@ -738,20 +781,21 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueueExt(HSAuint32 NodeId,
 		doorbell_offset = q->queue_id * DOORBELL_SIZE(q->gfxv);
 	}
 
-	err = map_doorbell(NodeId, gpu_id, doorbell_mmap_offset);
+	err = map_doorbell(ctx, NodeId, gpu_id, doorbell_mmap_offset);
 	if (err != HSAKMT_STATUS_SUCCESS) {
-		hsaKmtDestroyQueue(q->queue_id);
+		hsaKmtDestroyQueueCtx(ctx, q->queue_id);
 		return HSAKMT_STATUS_ERROR;
 	}
 
 	QueueResource->QueueId = PORT_VPTR_TO_UINT64(q);
-	QueueResource->Queue_DoorBell = VOID_PTR_ADD(doorbells[NodeId].mapping,
+	QueueResource->Queue_DoorBell = VOID_PTR_ADD(queue_ctx->doorbells[NodeId].mapping,
 						     doorbell_offset);
 
 	return HSAKMT_STATUS_SUCCESS;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtUpdateQueue(HSA_QUEUEID QueueId,
+HSAKMT_STATUS HSAKMTAPI hsaKmtUpdateQueueCtx(HsaKFDContext *ctx,
+					  HSA_QUEUEID QueueId,
 					  HSAuint32 QueuePercentage,
 					  HSA_QUEUE_PRIORITY Priority,
 					  void *QueueAddress,
@@ -775,7 +819,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtUpdateQueue(HSA_QUEUEID QueueId,
 	arg.queue_percentage = QueuePercentage;
 	arg.queue_priority = priority_map[Priority+3];
 
-	int err = hsakmt_ioctl(hsakmt_kfd_fd, AMDKFD_IOC_UPDATE_QUEUE, &arg);
+	int err = hsakmt_ioctl(ctx->fd, AMDKFD_IOC_UPDATE_QUEUE, &arg);
 
 	if (err == -1)
 		return HSAKMT_STATUS_ERROR;
@@ -783,7 +827,8 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtUpdateQueue(HSA_QUEUEID QueueId,
 	return HSAKMT_STATUS_SUCCESS;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtDestroyQueue(HSA_QUEUEID QueueId)
+HSAKMT_STATUS HSAKMTAPI hsaKmtDestroyQueueCtx(HsaKFDContext *ctx,
+						 HSA_QUEUEID QueueId)
 {
 	CHECK_KFD_OPEN();
 
@@ -795,20 +840,21 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtDestroyQueue(HSA_QUEUEID QueueId)
 
 	args.queue_id = q->queue_id;
 
-	int err = hsakmt_ioctl(hsakmt_kfd_fd, AMDKFD_IOC_DESTROY_QUEUE, &args);
+	int err = hsakmt_ioctl(ctx->fd, AMDKFD_IOC_DESTROY_QUEUE, &args);
 
 	if (err == -1) {
 		pr_err("Failed to destroy queue: %s\n", strerror(errno));
 		return HSAKMT_STATUS_ERROR;
 	}
 
-	free_queue(q);
+	free_queue(ctx, q);
 	return HSAKMT_STATUS_SUCCESS;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtSetQueueCUMask(HSA_QUEUEID QueueId,
-					     HSAuint32 CUMaskCount,
-					     HSAuint32 *QueueCUMask)
+HSAKMT_STATUS HSAKMTAPI hsaKmtSetQueueCUMaskCtx(HsaKFDContext *ctx,
+						 HSA_QUEUEID QueueId,
+						 HSAuint32 CUMaskCount,
+						 HSAuint32 *QueueCUMask)
 {
 	struct queue *q = PORT_UINT64_TO_VPTR(QueueId);
 	struct kfd_ioctl_set_cu_mask_args args = {0};
@@ -822,7 +868,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtSetQueueCUMask(HSA_QUEUEID QueueId,
 	args.num_cu_mask = CUMaskCount;
 	args.cu_mask_ptr = (uintptr_t)QueueCUMask;
 
-	int err = hsakmt_ioctl(hsakmt_kfd_fd, AMDKFD_IOC_SET_CU_MASK, &args);
+	int err = hsakmt_ioctl(ctx->fd, AMDKFD_IOC_SET_CU_MASK, &args);
 
 	if (err == -1)
 		return HSAKMT_STATUS_ERROR;
@@ -833,12 +879,9 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtSetQueueCUMask(HSA_QUEUEID QueueId,
 	return HSAKMT_STATUS_SUCCESS;
 }
 
-HSAKMT_STATUS
-HSAKMTAPI
-hsaKmtGetQueueInfo(
-	HSA_QUEUEID QueueId,
-	HsaQueueInfo *QueueInfo
-)
+HSAKMT_STATUS HSAKMTAPI hsaKmtGetQueueInfoCtx(HsaKFDContext *ctx,
+						 HSA_QUEUEID QueueId,
+						 HsaQueueInfo *QueueInfo)
 {
 	struct queue *q = PORT_UINT64_TO_VPTR(QueueId);
 	struct kfd_ioctl_get_queue_wave_state_args args = {0};
@@ -854,7 +897,7 @@ hsaKmtGetQueueInfo(
 	args.queue_id = q->queue_id;
 	args.ctl_stack_address = (uintptr_t)q->ctx_save_restore;
 
-	if (hsakmt_ioctl(hsakmt_kfd_fd, AMDKFD_IOC_GET_QUEUE_WAVE_STATE, &args) < 0)
+	if (hsakmt_ioctl(ctx->fd, AMDKFD_IOC_GET_QUEUE_WAVE_STATE, &args) < 0)
 		return HSAKMT_STATUS_ERROR;
 
 	QueueInfo->ControlStackTop = (void *)(args.ctl_stack_address +
@@ -872,7 +915,8 @@ hsaKmtGetQueueInfo(
 	return HSAKMT_STATUS_SUCCESS;
 }
 
-HSAKMT_STATUS HSAKMTAPI hsaKmtSetTrapHandler(HSAuint32 Node,
+HSAKMT_STATUS HSAKMTAPI hsaKmtSetTrapHandlerCtx(HsaKFDContext *ctx,
+						 HSAuint32 Node,
 					     void *TrapHandlerBaseAddress,
 					     HSAuint64 TrapHandlerSizeInBytes,
 					     void *TrapBufferBaseAddress,
@@ -892,7 +936,7 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtSetTrapHandler(HSAuint32 Node,
 	args.tba_addr = (uintptr_t)TrapHandlerBaseAddress;
 	args.tma_addr = (uintptr_t)TrapBufferBaseAddress;
 
-	int err = hsakmt_ioctl(hsakmt_kfd_fd, AMDKFD_IOC_SET_TRAP_HANDLER, &args);
+	int err = hsakmt_ioctl(ctx->fd, AMDKFD_IOC_SET_TRAP_HANDLER, &args);
 
 	return (err == -1) ? HSAKMT_STATUS_ERROR : HSAKMT_STATUS_SUCCESS;
 }
@@ -922,12 +966,10 @@ uint32_t *hsakmt_convert_queue_ids(HSAuint32 NumQueues, HSA_QUEUEID *Queues)
 	return queue_ids_ptr;
 }
 
-HSAKMT_STATUS
-HSAKMTAPI
-hsaKmtAllocQueueGWS(
-                HSA_QUEUEID        QueueId,
-                HSAuint32          nGWS,
-                HSAuint32          *firstGWS)
+HSAKMT_STATUS HSAKMTAPI hsaKmtAllocQueueGWSCtx(HsaKFDContext *ctx,
+						 HSA_QUEUEID QueueId,
+						 HSAuint32 nGWS,
+						 HSAuint32 *firstGWS)
 {
 	struct kfd_ioctl_alloc_queue_gws_args args = {0};
 	struct queue *q = PORT_UINT64_TO_VPTR(QueueId);
@@ -937,7 +979,7 @@ hsaKmtAllocQueueGWS(
 	args.queue_id = (HSAuint32)q->queue_id;
 	args.num_gws = nGWS;
 
-	int err = hsakmt_ioctl(hsakmt_kfd_fd, AMDKFD_IOC_ALLOC_QUEUE_GWS, &args);
+	int err = hsakmt_ioctl(ctx->fd, AMDKFD_IOC_ALLOC_QUEUE_GWS, &args);
 
 	if (!err && firstGWS)
 		*firstGWS = args.first_gws;
@@ -953,3 +995,85 @@ hsaKmtAllocQueueGWS(
 	else
 		return HSAKMT_STATUS_ERROR;
 }
+
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueue(HSAuint32 NodeId,
+						 HSA_QUEUE_TYPE Type,
+						 HSAuint32 QueuePercentage,
+						 HSA_QUEUE_PRIORITY Priority,
+						 void *QueueAddress,
+						 HSAuint64 QueueSizeInBytes,
+						 HsaEvent *Event,
+						 HsaQueueResource *QueueResource)
+{
+	if (Type == HSA_QUEUE_SDMA_BY_ENG_ID)
+		return HSAKMT_STATUS_ERROR;
+
+	return hsaKmtCreateQueueExt(NodeId, Type, QueuePercentage, Priority, 0,
+				    QueueAddress, QueueSizeInBytes, Event,
+				    QueueResource);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueueExt(HSAuint32 NodeId,
+					     HSA_QUEUE_TYPE Type,
+					     HSAuint32 QueuePercentage,
+					     HSA_QUEUE_PRIORITY Priority,
+					     HSAuint32 SdmaEngineId,
+					     void *QueueAddress,
+					     HSAuint64 QueueSizeInBytes,
+					     HsaEvent *Event,
+					     HsaQueueResource *QueueResource)
+{
+
+	return hsaKmtCreateQueueExtCtx(&hsakmt_primary_kfd_ctx, NodeId, Type,
+					QueuePercentage, Priority, SdmaEngineId, QueueAddress,
+					QueueSizeInBytes, Event, QueueResource);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtUpdateQueue(HSA_QUEUEID QueueId,
+						 HSAuint32 QueuePercentage,
+						 HSA_QUEUE_PRIORITY Priority,
+						 void *QueueAddress,
+						 HSAuint64 QueueSize,
+						 HsaEvent *Event)
+{
+	return hsaKmtUpdateQueueCtx(&hsakmt_primary_kfd_ctx, QueueId, QueuePercentage,
+					Priority, QueueAddress, QueueSize, Event);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtDestroyQueue(HSA_QUEUEID QueueId)
+{
+	return hsaKmtDestroyQueueCtx(&hsakmt_primary_kfd_ctx, QueueId);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtSetQueueCUMask(HSA_QUEUEID QueueId,
+						 HSAuint32 CUMaskCount,
+						 HSAuint32 *QueueCUMask)
+{
+	return hsaKmtSetQueueCUMaskCtx(&hsakmt_primary_kfd_ctx, QueueId, CUMaskCount, QueueCUMask);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtGetQueueInfo(
+						 HSA_QUEUEID QueueId,
+						 HsaQueueInfo *QueueInfo)
+{
+	return hsaKmtGetQueueInfoCtx(&hsakmt_primary_kfd_ctx, QueueId, QueueInfo);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtSetTrapHandler(HSAuint32 Node,
+						 void *TrapHandlerBaseAddress,
+						 HSAuint64 TrapHandlerSizeInBytes,
+						 void *TrapBufferBaseAddress,
+						 HSAuint64 TrapBufferSizeInBytes)
+{
+	return hsaKmtSetTrapHandlerCtx(&hsakmt_primary_kfd_ctx, Node,
+					TrapHandlerBaseAddress, TrapHandlerSizeInBytes,
+					TrapBufferBaseAddress, TrapBufferSizeInBytes);
+}
+
+HSAKMT_STATUS HSAKMTAPI hsaKmtAllocQueueGWS(HSA_QUEUEID QueueId,
+						 HSAuint32 nGWS,
+						 HSAuint32 *firstGWS)
+{
+	return hsaKmtAllocQueueGWSCtx(&hsakmt_primary_kfd_ctx, QueueId, nGWS, firstGWS);
+}
