@@ -41,11 +41,6 @@
 ////////////////////////////////////////////////////////////////////////////////
 
 #include "blit_kernel.h"
-
-#if (defined(WIN32) || defined(_WIN32))
-#define NOMINMAX
-#endif
-
 #include <algorithm>
 #include <atomic>
 #include <sstream>
